/*
	$Id:  $

	Copyright (C) Nicklas Nordborg
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.servlet;

import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.sf.basedb.clients.web.extensions.ExtensionsControl;
import net.sf.basedb.clients.web.extensions.RequestWrapper;
import net.sf.basedb.clients.web.extensions.ServletWrapper;
import net.sf.basedb.clients.web.extensions.ServletWrapperKey;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.manager.ExtensionsFile;

/**
	This servlet is acting as a proxy for requests to extension defined
	servlets. Such servlets are defined in <code>META-INF/servlets.xml</code>
	in the extension JAR file. A request to a servlet must have the following
	URL pattern:
	<code>/extensions/[jar-name]/[servlet-name].servlet</code>
	
	@author nicklas
	@version 2.7
	@base.modified $Date: 2008-03-17 14:40:46 +0100 (Mon, 17 Mar 2008) $
*/
public class ExtensionsServlet 
	extends HttpServlet
{
	
	private static final long serialVersionUID = 338904698493055683L;

	public ExtensionsServlet()
	{}
	
	/**
		Path to servlet must match: 
		/extensions/[jar-file-name]/[servlet-name]/optional/path/info.servlet
		group(1) = Servlet path; group(2) = JAR name, group(3) = Servlet name
		group(4) = Path info (not including .servlet)
	*/
	private static final Pattern PATH_MATCH = 
		Pattern.compile("(" + ExtensionsControl.RESOURCES_URL + "/([^/]+)/([^/]+))(/?.*)\\.servlet");
	
	/**
		Alternatively the servlet is: /extensions/servlet
		and pathInfo is: /[jar-file-name]/[servlet-name]/optional/path/info
	 */
	private static final Pattern ALTERNATE_PATH_MATCH = 
		Pattern.compile("(" + ExtensionsControl.RESOURCES_URL + "/servlet/([^/]+)/([^/]+))(/?.*)");
		
	
	@Override
	protected void doGet(HttpServletRequest request, HttpServletResponse response)
		throws ServletException, IOException
	{
		String servletPath = request.getServletPath();
		Pattern pathPattern = PATH_MATCH;
		if (!servletPath.endsWith(".servlet"))
		{
			servletPath += request.getPathInfo();
			pathPattern = ALTERNATE_PATH_MATCH;
		}
		
		Matcher m = pathPattern.matcher(servletPath);
		if (!m.matches())
		{
			response.sendError(HttpServletResponse.SC_NOT_FOUND, servletPath);
			return;
		}
		
		final String extensionServletPath = m.group(1);
		final String jarName = m.group(2);
		final String servletName = m.group(3);
		final String pathInfo = Values.getStringOrNull(m.group(4));

		ExtensionsControl ec = ExtensionsControl.get(null);
		ServletWrapperKey servletKey = new ServletWrapperKey(jarName, servletName);
		ExtensionsFile file = ec.getFileByObjectKey(servletKey);
		if (file == null)
		{
			response.sendError(HttpServletResponse.SC_NOT_FOUND, servletPath);
			return;
		}
		else if (!file.isJar())
		{
			response.sendError(HttpServletResponse.SC_BAD_REQUEST, servletPath + " is not a JAR file.");
			return;
		}
	
		ServletWrapper wrapper = file.getObjectForKey(servletKey);
		if (wrapper == null)
		{
			response.sendError(HttpServletResponse.SC_NOT_FOUND, servletPath);
			return;
		}

		wrapper.service(new RequestWrapper(request, extensionServletPath, pathInfo), response);
	}

	@Override
	protected void doPost(HttpServletRequest req, HttpServletResponse resp)
		throws ServletException, IOException
	{
		doGet(req, resp);
	}

}
