/*
	$Id: AbstractEntityQuery.java 7007 2015-11-11 10:35:21Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.QueryType;

import java.util.Collection;
import java.util.EnumSet;
import java.util.List;

/**
	An abstract implementation of the {@link EntityQuery} interface. This class
	doesn't support adding selections, groups or having restrictions.
	<p>
	{@link Include} options and permission control are implemented using 
	Hibernate filters. See {@link QueryRuntimeFilterFactory}.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-11-11 11:35:21 +0100 (on, 11 nov 2015) $
*/
abstract class AbstractEntityQuery
	extends AbstractHqlQuery
	implements EntityQuery
{

	/**
		Log all parameter bindings to prepared statement.
	*/
	private static final org.slf4j.Logger logParam = 
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.query.bind.AbstractEntityQuery");
		
	/**
		So we don't always have to call logParam.debug()
	*/
	private static final boolean debugEnabled = logParam.isDebugEnabled();

	/**
		The item type returned by the query, eg.
		SELECT [returnType] FROM ...
	*/
	private final Item returnType;
	
	/**
		The root item type we start with in the query. eq.
		SELECT ... FROM [rootTypeTable]
		@since 2.8
	*/
	private final Item rootType;
	
	/**
		The required runtime filter.
	*/
	private final QueryRuntimeFilter requiredFilter;
	
	/**
		Optional runtime filter.
	*/
	private final QueryRuntimeFilter optionalFilter;

	/**
		The manager of the runtime query filters.
	*/
	private QueryRuntimeFilterManager filterManager;
	
	/**
		Flags specifying which items to include.
	*/
	private EnumSet<Include> includes;
	
	/**
		The logged in user's permission on the items.
	*/
	private Permission permission;
	
	/**
		Create a new query returning items of the specified item type
		using the default optional runtime filter.
		@param returnType The type of items that are returned
		@param entityName The name of the type of items as known to
			Hibernate, or null to use the class of the return type
		@param stateless TRUE if the stateless Hibernate session should be used, FALSE
			if the regular Hibernate session should be used
	*/
	AbstractEntityQuery(Item returnType, String entityName, boolean stateless)
	{
		this(returnType, returnType, entityName, null, stateless, 
				QueryRuntimeFilterFactory.getOptionalFilter(returnType));
	}
	
	/**
		Create a new query returning items of the specified type with a non-default
		optional runtime filter.
		@param returnType The type of items that are returned
		@param entityName The name of the type of items as known to
			Hibernate, or null to use the class of the return type
		@param stateless TRUE if the stateless Hibernate session should be used, FALSE
			if the regular Hibernate session should be used
		@param optionalFilter A runtime filter replacing the default optional filter
			or null to not use any optional filter (not supported for stateless sessions)
	*/
	AbstractEntityQuery(Item returnType, String entityName, boolean stateless, QueryRuntimeFilter optionalFilter)
	{
		this(returnType, returnType, entityName, null, stateless, optionalFilter);
	}

	/**
		Create a new query that may have different return type and root item type.

		@see DataQuery#DataQuery(Class, Item, String, String)
		@since 2.8
	*/
	AbstractEntityQuery(Item returnType, Item rootType, String rootName, String select, 
			boolean stateless, QueryRuntimeFilter optionalFilter)
	{
		super(rootName == null ? rootType.getDataClass().getName() : rootName, rootType.getAlias(), stateless);
		super.selectPermanent(Selects.expression(Hql.property(rootType.getAlias(), select), null, true));
		this.returnType = returnType;
		this.rootType = rootType;
		this.requiredFilter = stateless ? null : QueryRuntimeFilterFactory.getRequiredFilter(returnType);
		this.optionalFilter = stateless ? null : optionalFilter;
		this.includes = EnumSet.of(Include.NOT_REMOVED, Include.MINE, Include.IN_PROJECT, 
				Include.ANNOTATED, Include.NOT_ANNOTATED);
		this.permission = Permission.READ;
		setAutoJoinType(JoinType.LEFT);
	}
	
	/*
		From the Query interface
		-------------------------------------------
	*/
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	@Override
	public void select(Select select)
	{
		throw new UnsupportedOperationException();
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	@Override
	public void selectPermanent(Select select)
	{
		throw new UnsupportedOperationException();
	}
	
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	@Override
	public void group(Expression expression)
	{
		throw new UnsupportedOperationException();
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	@Override
	public void groupPermanent(Expression expression)
	{
		throw new UnsupportedOperationException();
	}

	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	@Override
	public void having(Restriction restriction)
	{
		throw new UnsupportedOperationException();
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	@Override
	public void havingPermanent(Restriction restriction)
	{
		throw new UnsupportedOperationException();
	}
	
	/**
		@return {@link QueryType#HQL}
	*/
	@Override
	public QueryType getQueryType()
	{
		return QueryType.HQL;
	}
	

	@Override
	public long count(DbControl dc)
		throws BaseException
	{
		long count = -1;
		enableFilters(dc);
		try
		{
			count = super.count(dc);
		}
		finally
		{
			disableFilters(dc);
		}
		return count;
	}
	// -------------------------------------------

	/*
		From the EntityQuery interface
		-------------------------------------------
	*/
	@Override
	public Item getItemType()
	{
		return returnType;
	}
	/**
		@since 2.8
	*/
	@Override
	public Item getRootType()
	{
		return rootType;
	}
	@Override
	public void include(Include... includes)
	{
		for (Include i : includes)
		{
			this.includes.add(i);
		}
	}
	@Override
	public void include(Collection<Include> includes)
	{
		this.includes.addAll(includes);
	}
	/**
	 	@since 2.14.2	 	
	*/
	@Override
	public void setIncludes(Collection<Include> includes)
	{
		this.includes.clear();
		this.includes.addAll(includes);
	}
	@Override
	public void exclude(Include... excludes)
	{
		for (Include i : excludes)
		{
			this.includes.remove(i);
		}
	}
	@Override
	public void exclude(Collection<Include> includes)
	{
		this.includes.removeAll(includes);
	}
	@Override
	public boolean isIncluded(Include... includes)
	{
		boolean isIncluded = true;
		for (Include i : includes)
		{
			isIncluded = isIncluded && this.includes.contains(i);
		}
		return isIncluded;
	}
	@Override
	public boolean isIncluded(Collection<Include> includes)
	{
		return this.includes.containsAll(includes);
	}
	@Override
	public void setItemPermission(Permission permission)
	{
		this.permission = permission == null ? Permission.READ : permission;
	}
	@Override
	public Permission getItemPermission()
	{
		return permission;
	}
	@Override
	public List<Integer> idList(DbControl dc)
	{
		enableFilters(dc);
		List<Integer> result = null;
		try
		{
			SessionControl sc = dc.getSessionControl();
			result = HibernateUtil.loadList(Integer.class, getIdHqlQuery(dc), sc);
		}
		finally
		{
			disableFilters(dc);
		}
		return result;
	}
	// -------------------------------------------
	
	/**
		Specify the join type of automatic joins. The default join type is
		{@link JoinType#LEFT}. This setting doesn't affect joins that has been made
		directly to the query.
		@param joinType The type of the automatic joins
		@since 2.8
	*/
	public void setAutoJoinType(JoinType joinType)
	{
		addAutoJoiner(HqlQueryJoiners.getReporterListJoiner(joinType));
		addAutoJoiner(HqlQueryJoiners.getAnnotationJoiner(joinType));
	}
	
	/**
		Enable runtime query filters for the query.
		@see #disableFilters(DbControl)
	*/
	void enableFilters(DbControl dc)
	{
		if (!isStateless() && (requiredFilter != null || optionalFilter != null))
		{
			filterManager = new QueryRuntimeFilterManager(dc, getItemType());
			if (requiredFilter != null) requiredFilter.enableFilters(filterManager, this, dc);
			if (optionalFilter != null) optionalFilter.enableFilters(filterManager, this, dc);
		}
	}
	
	/**
		Disable all enabled runtime query filters for the query.
		@see #enableFilters(DbControl)
	*/
	void disableFilters(DbControl dc)
	{
		if (filterManager != null)
		{
			filterManager.disableAll();
			filterManager = null;
		}
	}

}
