/*
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.RootRawBioAssayData;
import net.sf.basedb.core.data.SharedData;

/**
	Represents the use of a raw bioassay in an experiment. Except
	for keeping the link between the experiment and raw bioassay 
	the main purpose of this class is to make it possible to store
	experimental factor values as annotations (primary or inherited)
	without "contaminating" the raw bioassay and without problems
	with access permissions. Access permissions for this item is
	follows that of the experiment it belongs to.
	
	@author nicklas
	@since 3.6
*/
public class RootRawBioAssay 
	extends ChildItem
	implements Annotatable, Nameable
{

	
	/**
		The type of item represented by this class.
		@see Item#ROOTRAWBIOASSAY
		@see #getType()
	*/
	public static final Item TYPE = Item.ROOTRAWBIOASSAY;

	/**
		Get a <code>RootRawBioAssay</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>RootRawBioAssay</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static RootRawBioAssay getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		RootRawBioAssay root = dc.loadItem(RootRawBioAssay.class, id);
		if (root == null) throw new ItemNotFoundException("RootRawBioAssay[id=" + id + "]");
		return root;
	}

	/**
		Get a query configured to retrieve raw bioassays.
		@return An {@link ItemQuery} object
	*/
	static ItemQuery<RootRawBioAssay> getQuery() 
	{
		return new ItemQuery<RootRawBioAssay>(RootRawBioAssay.class, null);
	}
	
	/**
	 	Creates a new root raw bioassay item from the given data.
		@param data the data
	*/
	RootRawBioAssay(RootRawBioAssayData data)
	{
		super(data);
	}
	
	@Override
	RootRawBioAssayData getData()
	{
		return (RootRawBioAssayData)getBasicData();
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	@Override
	public AnnotationSet getAnnotationSet()
		throws PermissionDeniedException, BaseException
	{
		DbControl dc = getDbControl();
		AnnotationSet as = dc.getItem(AnnotationSet.class, getData().getAnnotationSet(), this);
		if (as == null)
		{
			as = AnnotationSet.getNew(dc, this);
			getData().setAnnotationSet(as.getData());
			dc.saveItemIf(this, as, false);
		}
		return as;
	}
	@Override
	public boolean isAnnotated()
	{
		return getData().getAnnotationSet() != null;
	}
	@Override
	public void removeAnnotations()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		AnnotationSetData as = getData().getAnnotationSet();
		if (as != null)
		{
			HibernateUtil.deleteData(getDbControl().getHibernateSession(), as);
			getData().setAnnotationSet(null);
		}
	}
	@Override
	public Protocol getProtocol()
	{
		return null;
	}
	/**
		Get the raw bioassay.
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		DbControl dc = getDbControl();
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		annotatable.add(getRawBioAssay());
		return annotatable;
	}
	// -------------------------------------------
	
	/*
		From BasicItem
		-------------------------------------------
	*/
	/**
		Deny DELETE permission. Root raw bioAssays are automatically 
		deleted when the parent experiment is deleted or by 
		calling {@link Experiment#removeRawBioAssay(RawBioAssay)}
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		super.initPermissions(granted, denied | Permission.deny(Permission.DELETE));
	}
	// -------------------------------------------

	/*
		From the ChildItem class
		-------------------------------------------
	*/
	/**
		Get the experiment.
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getExperiment();
	}
	/**
		WRITE permission is required to be able to manage root raw bioassays.
	*/
	@Override
	Permission getPermissionForWriteDeleteAndCreate()
	{
		return Permission.WRITE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public String getName()
	{
		return getData().getName();
	}
	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	@Override
	public String getDescription()
	{
		return getData().getDescription();
	}
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------

	/**
		Get the experiment this root raw bioassay belongs to.
		@return An <code>Experiment</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Experiment getExperiment()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Experiment.class, getData().getExperiment());
	}
	
	/**
		Get the raw bioassay this root raw bioassay is representing.
		@return A <code>RawBioAssay</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the bioassayset
		@throws BaseException If there is another error
	*/
	public RawBioAssay getRawBioAssay()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(RawBioAssay.class, getData().getRawBioAssay());
	}


}
