/**
	$Id: SignalTransporter.java 4516 2008-09-11 20:09:17Z jari $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.signal;

import java.util.Collection;

/**
	A signal transporter is a class that knows how to send a signal to a
	specific {@link SignalReceiver} implementation. Implementing classes
	must have a public no-argument constructor. Before signals can be sent
	with the transporter the {@link #init(String)} method is called with the
	same string as returned by the {@link SignalReceiver#registerSignalHandler(SignalHandler)}
	method. This string should contain enough information to let the transporter
	send the signal to the correct signal handler on the correct signal receiver.
	The format of this string is implementation-specific and needs only to be known
	by the coupled transporter and receiver classes. As an example, the {@link SocketSignalReceiver}
	and {@link SocketSignalTransporter} uses a {@link java.net.URI} that contains
	the IP number and port of the receiver and an ID for identifying the signal handler.
	<p>
	Signal transporters need not be thread safe since a new instance is created 
	for each signal that is going to be sent.

	@author nicklas
	@version 2.6
	@base.modified $Date: 2008-09-11 22:09:17 +0200 (to, 11 sep 2008) $
*/
public interface SignalTransporter
{
	/**
		Initialise the signal transporter. The parameter is the same as
		returned by {@link SignalReceiver#registerSignalHandler(SignalHandler)}
		and should contain information that the transporter can parse and use for 
		contacting the correct signal receiver and to send signals to the
		registered signal handler.
		
		@param params The initialisation string
	*/
	public void init(String params);
	
	/**
		Get a collection containing the signals supported, by the signal
		handler. A null return value means that the transporter just doesn't
		know about which signals are supported and which are not. 
		@return A collections with the signals, or null
	*/
	public Collection<Signal> getSupportedSignals();
	
	/**
		Send a signal. 
		@param signal The signal to send, must not be null
	*/
	public void send(Signal signal);
}
