/*
	$Id: FormulaFilter.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg, Martin Svensson
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import net.sf.basedb.core.ProgressReporter;


import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.plugin.AbstractAnalysisPlugin;
import net.sf.basedb.core.plugin.AnalysisFilterPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.BioAssaySetFilterUtil;
import net.sf.basedb.util.BioAssaySetUtil;
import net.sf.basedb.util.DynamicFilter;
import net.sf.basedb.util.IncludeExcludeFilter;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
   This plugin provides methods to filter a bioassayset based on 
   formulas entered by the user.

   @author Nicklas
   @version 2.0
   @base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class FormulaFilter
	extends AbstractAnalysisPlugin
	implements InteractivePlugin, AnalysisFilterPlugin, SignalTarget
{


	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	private static final Set<GuiContext> guiContexts = Collections.unmodifiableSet(
		new HashSet<GuiContext>(Arrays.asList(CONTEXT_BIOASSAYSET, CONTEXT_BIOASSAYS))
	);

	private RequestInformation configureJob;

	private ThreadSignalHandler signalHandler;
	
	/**
	   Constructor should be empty. 
	*/
	public FormulaFilter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request use access to Experiment:s and read access to Job:s and ReporterLists.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.EXPERIMENT, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.JOB, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.REPORTERLIST, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}

	@Override
	public void run(Request request, Response response, ProgressReporter progress)
	{
		DbControl dc = sc.newDbControl();
		try
		{
			if (signalHandler != null) 
			{
				signalHandler.setWorkerThread(null);
			}

			// Get all parameters
			BioAssaySet source = getSourceBioAssaySet(dc);
			List<BioAssay> bioAssays = getSourceBioAssays(dc);
			String expression = (String)job.getValue("expression");
			Integer includeLimit = (Integer)job.getValue("includeLimit");
			Integer excludeLimit = (Integer)job.getValue("excludeLimit");
			String childName = (String)job.getValue("childName");
			String childDescription = (String)job.getValue("childDescription");
			
			// get the current job
			Job currentJob = getCurrentJob(dc);
			
			// The bioassay set to create
			BioAssaySet child = null;
			
			// Parse filter expression
			// NOTE! Here disobey the advice to not use the restriction on a different query
			// but it should be safe since an equivalent query is used by the
			// BioAssaySetFilterUtil.createFilteredBioAssaySet() method
			Restriction filter = BioAssaySetUtil.createJepRestriction(dc, expression, source.getSpotData());
			
			ThreadSignalHandler.checkInterrupted();

			if (includeLimit != null || excludeLimit != null)
			{
				DynamicFilter dFilter = 
					new IncludeExcludeFilter(dc, source, bioAssays, filter, includeLimit, excludeLimit);
				child = BioAssaySetFilterUtil.createFilteredBioAssaySet(dc, source, bioAssays, 
						currentJob, dFilter, progress);
			}
			else
			{
				child = BioAssaySetFilterUtil.createFilteredBioAssaySet(dc, source, bioAssays,
						currentJob, filter, progress);
			}

			if (expression.length() > 240) expression = expression.substring(0, 240) + "..."; 
			child.getTransformation().setName("Filter: " + expression);
			child.setName(childName);
			child.setDescription(childDescription);
			dc.commit();

			int sourceSpots = source.getNumSpots();
			if (bioAssays != null && bioAssays.size() > 0)
			{
				sourceSpots = 0;
				for (BioAssay ba : bioAssays)
				{
					sourceSpots += ba.getNumSpots();
				}
			}
			int childSpots = child.getNumSpots();
			response.setDone(childSpots +" spots remaining; " + (sourceSpots - childSpots) + " removed.");
		}
		catch (Throwable t)
		{
			response.setError(t.getMessage(), Arrays.asList(t));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// -------------------------------------------
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = super.isInContext(context, item);
		if (message == null)
		{
			BioAssaySet bas = null;
			if (context.equals(CONTEXT_BIOASSAYSET))
			{
				bas = (BioAssaySet)item;
			}
			else
			{
				DbControl dc = sc.newDbControl();
				try
				{
					bas = getCurrentBioAssaySet(dc);
				}
				finally
				{
					if (dc != null) dc.close();
				}
			}
			if (bas.getNumSpots() <= 0)
			{
				message = "This plug-in requires spot data in the database";
			}
		}
		return message;
	}
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
		}
		return requestInformation;
	}
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				
				// Source bioassay set and bioassays
				storeValue(job, request, ri.getParameter(SOURCE_BIOASSAYSET));
				storeValues(job, request, ri.getParameter(SOURCE_BIOASSAYS));
				
				// Filter expression and options
				storeValue(job, request, ri.getParameter("expression"));
				storeValue(job, request, ri.getParameter("includeLimit"));
				storeValue(job, request, ri.getParameter("excludeLimit"));
				
				// Child name and description
				storeValue(job, request, ri.getParameter(CHILD_NAME));
				storeValue(job, request, ri.getParameter(CHILD_DESCRIPTION));
				
				BioAssaySet bioAssaySet = (BioAssaySet)job.getValue(SOURCE_BIOASSAYSET);
				response.setSuggestedJobName("Filter bioassay set '" + bioAssaySet.getName() + 
						"': " + job.getValue("expression"));
				
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	
	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		DbControl dc = null;
		try
		{
			if (configureJob == null)
			{
				// Get current bio assay set
				dc = sc.newDbControl();
				BioAssaySet bas = getCurrentBioAssaySet(dc);				
				
				// Parameters for CONFIGURE_JOB
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

				// Source bioassay set and bioassay parameters
				parameters.add(getSourceBioAssaySetParameter(null, null));
				parameters.add(getSourceBioAssaysParameter(null, null));
				
				// Filter expression and options
				parameters.add(new PluginParameter<String>(
					"expression", "Filter expression", 
					"The expression used to filter the data",
					new StringParameterType(255, null, true)
				));
				IntegerParameterType limitParameterType = new IntegerParameterType(0, null, null, false);
				parameters.add(new PluginParameter<Integer>(
					"includeLimit", "Keep non-matching spots", 
					"Keep non-matching spots from all bioassays if the filter matches that spot in at least this number of bioassays",
					limitParameterType
				));
				parameters.add(new PluginParameter<Integer>(
					"excludeLimit", "Exclude matching spots", 
					"Exclude matching spots from all bioassays if the filter doesn't match that spot in at least this number of bioassays",
					limitParameterType
				));
				
				// Child bioassay set name and description
				parameters.add(getChildNameParameter(null, null, bas.getName()));
				parameters.add(getChildDescriptionParameter(null, null, null));
				
				configureJob = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_JOB,
					"Specify filter expression",
					"You should enter a filter expression and set other options for the filter.",
					parameters, "jep_filter.jsp"
				);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}
	
}
