/**
	$Id: GalExporter.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.ArrayDesignBlock;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DataQuery;
import net.sf.basedb.core.DataResultIterator;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Feature;
import net.sf.basedb.core.FeatureCoordinate;
import net.sf.basedb.core.FeatureIdentificationMethod;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.Job.ExecutionTime;
import net.sf.basedb.core.data.FeatureData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.plugin.AbstractExporterPlugin;
import net.sf.basedb.core.plugin.ExportOutputStream;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;

/**
	Exports an array design with features to a GAL file. The exporter only
	works for array desigs that has stored the features in a database. For
	a description of the GAL format see 
	http://www.moleculardevices.com/pages/software/gn_genepix_file_formats.html#gal

	@author nicklas
	@version 2.7
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class GalExporter
	extends AbstractExporterPlugin
	implements SignalTarget, InteractivePlugin
{
	
	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.ARRAYDESIGN, GuiContext.Type.ITEM));
	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	private ThreadSignalHandler signalHandler;
	private RequestInformation configureJob;

	public GalExporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request create and write access to File:s, Directory:s and ArrayDesign:s
	 	read access to User:s and forced read access to DataFileType:s.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			synchronized(permissions)
			{
				permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
				permissions.add(new Permissions(Item.DIRECTORY, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
				permissions.add(new Permissions(Item.ARRAYDESIGN, null, EnumSet.of(Permission.WRITE)));
				permissions.add(new Permissions(Item.USER, null, EnumSet.of(Permission.READ)));
				permissions.add(new Permissions(Item.DATAFILETYPE, EnumSet.of(Permission.READ), null));
			}
		}
		return permissions;
	}
	// --------------------------------------------
	
	/*
	 	From the InteractivePlugin interface
	 	-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;
		if (item == null)
		{
			message = "The object is null";
		}
		else if (!(item instanceof ArrayDesign))
		{
			message = "The object is not a ArrayDesign: " + item;
		}
		else
		{
			ArrayDesign design = (ArrayDesign)item;
			if (design.getNumDbFeatures() == 0)
			{
				message = "The array design has no features in the database";
			}
			else if (design.getFeatureIdentificationMethod() != FeatureIdentificationMethod.COORDINATES)
			{
				message = "The array design must use COORDINATES as feature identification method.";
			}
		}
		return message;
	}
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (Request.COMMAND_CONFIGURE_JOB.equals(command))
		{
			requestInformation = getConfigureJobParameters(null);
		}
		return requestInformation;
	}
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();		
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(!request.isAllowedImmediateExecution());
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameters were found in the request", errors);
					return;
				}				
				
				storeValue(job, request, ri.getParameter("design"));
				storeValue(job, request, ri.getParameter("featureDiameter"));
				
				if (request.getParameterValue(SAVE_AS) == null)
				{
					if (!request.isAllowedImmediateExecution())
					{
						response.setError("Immediate download is not allowed. Please specify a filename.", null);
						return;
					}
					ArrayDesign design = (ArrayDesign)request.getParameterValue("design");
					response.setDownloadImmediately("Export feature in GAL format for " + 
						design.getName(), ExecutionTime.SHORTEST, true);
				}
				else
				{
					Object parameterValue = request.getParameterValue(OVERWRITE); 
					boolean overwrite = parameterValue != null ? (Boolean)parameterValue : false;
					if (!pathCanBeUsed((String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), overwrite))
					{
						response.setError("File exists: " + (String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), null);
						return;
					}
					storeValue(job, request, ri.getParameter(SAVE_AS));
					storeValue(job, request, ri.getParameter(OVERWRITE));
					storeValue(job, request, ri.getParameter("setAsReporterMap"));
					
					ArrayDesign design = (ArrayDesign)job.getValue("design");
					response.setSuggestedJobName("Export array design '" + design.getName() +
							"' as GAL file: " + job.getValue(SAVE_AS));
					
					response.setDone("The job configuration is complete", ExecutionTime.SHORTEST);
				}
			}
		}
		catch(Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}		
	}
	// -------------------------------------------

	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------
	
	/*
		From the AbstractExporterPlugin class
		-------------------------------------------
	*/
	private DbControl dc;
	private long numExported;
	@Override
	protected void begin(DbControl dc)
	{
		this.dc = dc;
	}
	
	@Override
	protected void performExport(ExportOutputStream out,
		ProgressReporter progress) throws IOException
	{
		ArrayDesign design = (ArrayDesign)job.getValue("design");
		Integer featureDiameter = (Integer)job.getValue("featureDiameter");
		design = ArrayDesign.getById(dc, design.getId());
		boolean setReporterMap = Boolean.TRUE.equals(job.getValue("setAsReporterMap"));
		out.setMimeType("text/plain");
		out.setFilename(Path.makeSafeFilename(design.getName(), "").toLowerCase() + ".gal");
		out.setCharacterSet("ISO-8859-1");
		PrintWriter writer = new PrintWriter(new OutputStreamWriter(out, "ISO-8859-1"));
		exportGal(dc, design, featureDiameter, writer, progress);
		writer.flush();
		out.flush();
		File saveTo = getSaveTo();
		if (setReporterMap && saveTo != null)
		{
			FileStoreUtil.setDataFile(dc, design, DataFileType.GENERIC_REPORTER_MAP, saveTo);
		}
	}
	@Override
	protected void end(boolean success)
	{
		this.dc = null;
	}
	@Override
	protected String getSuccessMessage()
	{
		return numExported + " features exported successfully";
	}
	// --------------------------------------------
	
	/**
		Exports an array design with features to a GAL file. The array
		must have features in the database.
		
		@param dc The DbControl to use for database access
		@param design The array design to export
		@param featureDiameter The expected feature diameter (in µm) of spots, 
			or null to use the values from the database
		@param writer The stream to write to
		@param progress An optional progress reporter
		@since 2.8
	*/
	public void exportGal(DbControl dc, ArrayDesign design, Integer featureDiameter,
		PrintWriter writer, ProgressReporter progress)
		throws IOException
	{
		// Validate parameters
		if (dc == null) throw new InvalidUseOfNullException("dc");
		if (design == null) throw new InvalidUseOfNullException("design");
		if (writer == null) throw new InvalidUseOfNullException("writer");
		if (design.getNumDbFeatures() == 0) return;

		if (signalHandler != null) 
		{
			signalHandler.setWorkerThread(null);
		}
		ThreadSignalHandler.checkInterrupted();
		if (progress != null) progress.display(0, "Loading blocks...");

		// Load the blocks
		ItemQuery<ArrayDesignBlock> blockQuery = design.getArrayDesignBlocks();
		blockQuery.order(Orders.asc(Hql.property("blockNumber")));
		List<ArrayDesignBlock> blocks = blockQuery.list(dc);
		
		// Fill a set with all possible coordinates so we can know which are missing
		// Using a LinkedHashSet to make the output appear in the insertion order
		Set<FeatureCoordinate> missing = new LinkedHashSet<FeatureCoordinate>(design.getNumDbFeatures());
		for (ArrayDesignBlock block : blocks)
		{
			int blockNum = block.getBlockNumber();
			int rows = block.getBlockSizeY();
			int columns = block.getBlockSizeX();
			for (int column = 1; column <= columns; ++column)
			{
				for (int row = 1; row <= rows; ++row)
				{
					missing.add(new FeatureCoordinate(blockNum, 0, 0, row, column));
				}
			}
		}
		
		// Create headers
		List<String> headers = new LinkedList<String>();
		createStandardHeaders(headers, design, blocks);
		createBlockHeaders(headers, design, featureDiameter, blocks);
		writeHeaders(writer, headers);

		// Start data output
		writeDataHeader(writer);
		ThreadSignalHandler.checkInterrupted();
		if (progress != null) progress.display(5, "Loading features...");
		
		// Query for loading features and their reporters
		DataQuery<FeatureData> featureQuery = design.getFeatures();
		// The fetch=true is important for performance
		featureQuery.join(Hql.leftJoin(null, "reporter", "rep", null, true));
		featureQuery.join(Hql.innerJoin(null, "arrayDesignBlock", "blk", true));
		featureQuery.order(Orders.asc(Hql.property("blk", "blockNumber")));
		featureQuery.order(Orders.asc(Hql.property("column")));
		featureQuery.order(Orders.asc(Hql.property("row")));

		DataResultIterator<FeatureData> result = featureQuery.iterate(dc);
		long total = design.getNumDbFeatures();
		long done = 0;
		long progressInterval = total / 50;
		if (progressInterval < 10) progressInterval = 10;
		
		while (result.hasNext())
		{
			FeatureData feature = result.next();
			ArrayDesignBlock block = Feature.getArrayDesignBlock(dc, feature);
			writeFeature(writer, feature, block);
			missing.remove(new FeatureCoordinate(block.getBlockNumber(), 0, 0, feature.getRow(), feature.getColumn()));
			done++;
			if (done % progressInterval == 0)
			{
				ThreadSignalHandler.checkInterrupted();
				if (progress != null)
				{
					progress.display((int)(5 + (95 * done) / total), 
							"Exporting features; " + done + " of " + total + " done.");
				}
			}
		}
		
		// Write missing features
		for (FeatureCoordinate feature : missing)
		{
			writeMissingFeature(writer, feature);
		}
		writer.flush();
		
		if (progress != null) progress.display(100, total + " features exported");
		numExported = total;
	}
	
	/**
		Create the standard headers that are required by the GAL format or that
		are useful to have in the GAL file.
		
		@param headers Add the created headers to this list; use 
			{@link #createHeader(String, Object)} to create a properly formatted header
		@param design The current array design
		@param blocks A list with all blocks in the array design
	*/
	protected void createStandardHeaders(List<String> headers, ArrayDesign design, List<ArrayDesignBlock> blocks)
	{
		headers.add(createHeader("Type", "GenePix ArrayList V1.0"));
		headers.add(createHeader("ArrayName", design.getName()));
		headers.add(createHeader("BlockCount", blocks.size()));
	}

	/**
		Create header entries for each block in the array design.
		@param headers Add the created headers to this list; use 
			{@link #createHeader(String, Object)} to create a properly formatted header
		@param design The current array design
		@param featureDiameter The expected feature diameter (µm), or null to
			use {@link ArrayDesignBlock#getFeatureDiameter()}
		@param blocks A list with all blocks in the array design
	*/
	protected void createBlockHeaders(List<String> headers, ArrayDesign design, 
		Integer featureDiameter, List<ArrayDesignBlock> blocks)
	{
		for (ArrayDesignBlock block : blocks)
		{
			int blockNum = block.getBlockNumber();
			StringBuilder blockData = new StringBuilder();
			blockData.append(block.getOriginX()).append(", ");
			blockData.append(block.getOriginY()).append(", ");
			if (featureDiameter == null)
			{
				blockData.append(block.getFeatureDiameter()).append(", ");
			}
			else
			{
				blockData.append(featureDiameter).append(", ");
			}
			blockData.append(block.getBlockSizeX()).append(", ");
			blockData.append(block.getSpacingX()).append(", ");
			blockData.append(block.getBlockSizeY()).append(", ");
			blockData.append(block.getSpacingY());
			headers.add(createHeader("Block" + blockNum, blockData.toString()));
		}	
	}
	
	/**
		Create a properly formatted GAL header. The typical format
		is "name=value". We assume that the header name has no strange
		characters in it, but use {@link #makeSafeHeaderValue(String)}
		for the value.
		
		@param name The name of the header
		@param value The value of the header
		@return A properly formatted header
	*/
	protected String createHeader(String name, Object value)
	{
		return "\"" + name + "=" + (value == null ? "" : makeSafeHeaderValue(value.toString())) + "\"";
	}
	
	private static final Pattern unsafe = Pattern.compile("\\\"|\n|\r");
	
	/**
		Removes "unsafe" characters in the value. Unsafe characers are:
		<ul>
		<li>quote (")
		<li>newline (\n, \r)
		</ul>
		
		@param value The value to check
		@return The checked value
	*/
	protected String makeSafeHeaderValue(String value)
	{
		return unsafe.matcher(value).replaceAll("");
	}
	
	/**
		Write the headers of the GAL file, including the two first lines
		with the "ATF 1.0" line and the second line with the number of header
		lines and the number of data columns. 
		
		@param writer The writer to print to
		@param headers The list with all headers
	*/
	protected void writeHeaders(PrintWriter writer, List<String> headers)
	{
		writer.println("ATF\t1.0");
		writer.println(headers.size() + "\t5");
		
		for (String header : headers)
		{
			writer.println(header);
		}
	}
	
	/**
		Write the data header. The column headers must match the data
		written by {@link #writeFeature(PrintWriter, FeatureData, ArrayDesignBlock)}
		and {@link #writeMissingFeature(PrintWriter, FeatureCoordinate)}.
		
		@param writer The writer to print to 
	*/
	protected void writeDataHeader(PrintWriter writer)
	{
		writer.println("\"Block\"\t\"Column\"\t\"Row\"\t\"ID\"\t\"Name\"");
	}
	
	/**
		Write feature data. The data that is written must match the column
		headers written by {@link #writeDataHeader(PrintWriter)}. Features
		that doesn't have a reporter should get ID and Name = empty.

		@param writer The writer to print to
		@param feature The feature
		@param block The block on which the feature is located
	*/
	protected void writeFeature(PrintWriter writer, FeatureData feature, ArrayDesignBlock block)
	{
		ReporterData reporter = feature.getReporter();
		writer.print(block.getBlockNumber());
		writer.print("\t");
		writer.print(feature.getColumn());
		writer.print("\t");
		writer.print(feature.getRow());
		writer.print("\t");
		if (reporter != null)
		{
			writer.print(reporter.getExternalId());
			writer.print("\t");
			writer.println(reporter.getName());
		}
		else
		{
			writer.println("empty\tempty");
		}
	}
	
	/**
		Write a missing features. The data that is written must match the column
		headers written by {@link #writeDataHeader(PrintWriter)}. The ID and Name
		should be empty.
	
		@param writer The writer to print to
		@param missing The feature
	*/
	protected void writeMissingFeature(PrintWriter writer, FeatureCoordinate missing)
	{
		writer.print(missing.getBlock());
		writer.print("\t");
		writer.print(missing.getColumn());
		writer.print("\t");
		writer.print(missing.getRow());
		writer.println("\tempty\tempty");
	}
	
	private RequestInformation getConfigureJobParameters(Boolean requireFile)
	{
		if (configureJob == null)
		{
			ParameterType<ArrayDesign> designType = 
				new ItemParameterType<ArrayDesign>(ArrayDesign.class, null, true, 1, null);
			PluginParameter<ArrayDesign> designParameter = new PluginParameter<ArrayDesign>(
					"design", "Array design", "The array design to export features from", designType);
			
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			parameters.add(designParameter);
			
			PluginParameter<Integer> featureDiameterParameter = 
				new PluginParameter<Integer>(
						"featureDiameter",
						"Override feature diameter",
						"Enter a value for the expected diameter (µm) of spots, or leave it empty to use " +
						"the values from the database.", 
						new IntegerParameterType()
					);
			
			parameters.add(featureDiameterParameter);
			if (requireFile == null)
			{
				PluginDefinition pd = job.getPluginDefinition();
				requireFile = pd == null ? false : !pd.getAllowImmediateExecution();
			}
			
			String defaultPath = null;
			ArrayDesign design = null;
			int currentDesignId = sc.getCurrentContext(Item.ARRAYDESIGN).getId();
			if (currentDesignId != 0)
			{
				DbControl dc = sc.newDbControl();
				try
				{
					design = ArrayDesign.getById(dc, currentDesignId);
					defaultPath = "~/" + Path.makeSafeFilename(design.getName(), "").toLowerCase() + ".gal";

					if (design != null && design.hasPermission(Permission.WRITE))
					{
						boolean hasReporterMap = FileStoreUtil.hasDataFile(dc, design, 
							DataFileType.GENERIC_REPORTER_MAP, false);
						PluginParameter<Boolean> setReporterMapParameter = new PluginParameter<Boolean>(
								"setAsReporterMap", "Set as reporter map", 
								"If this option is selected, the exported file will be set as the " +
								"'reporter map' file of the array design. This option is ignored if " +
								"no filename has been specified.", 
								new BooleanParameterType(!hasReporterMap, false));
						
						parameters.add(setReporterMapParameter);
					}				
				}
				finally
				{
					if (dc != null) dc.close();
				}
			}
			

			parameters.add(getSaveAsParameter(null, null, defaultPath, requireFile));
			parameters.add(getOverwriteParameter(null, null));
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"GAL file export options",
				"Set export options for the GAL file.",
				parameters
			);
		}
		return configureJob;
	}	
	
}
